from response_mapping import response_Brute_force_pipeline
from theme_extraction import themes_Brute_force_pipeline
import yaml
import argparse
import LLMs

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, default="")
    args = parser.parse_args()
    if args.config=="":
        raise ValueError('Please provide a path to your config file using python Brute_force.py --config your_config_file_path')
    ### model configuration
    with open(args.config, 'r') as file:
        config = yaml.safe_load(file)
    demo = dict(config['DEMO'])
    models = config['MODELS']
    consultation_name = config['CONSULTATIONS'][0]
    for model_name in models:
        #for consultation_name in consultation_name s:
        print(f"Analyzing {consultation_name} with model {model_name}")
        #response_Brute_force_pipeline(demo, consultation_name, model_name)
        themes_Brute_force_pipeline(demo, consultation_name, model_name)
        print("Brute force pipeline completed successfully for consultation: ", consultation_name)



# Import basic necessary packages
from sentence_transformers import SentenceTransformer
from transformers import AutoTokenizer
from timeit import default_timer as timer
from transformers import LlamaTokenizer
from transformers import GPT2Tokenizer
from transformers import T5Tokenizer
tokenizer = GPT2Tokenizer. from_pretrained("gpt2") ### select tokenizer
import tiktoken
tokenizer = tiktoken.encoding_for_model('gpt-3.5-turbo')

def calculate_chunks_theme_extraction(responses, system_prompt, model):
    context_length = model.context_length
    per_token_input_cost = model.input_cost
    per_token_output_cost = model.output_cost
    if (system_prompt == ""):
        system_prompt = """Instructions: 
        1. Identify the main theme(s) in the open-ended provided responses that explain the respondents' multiple choice selection. 
        2. Provide detailed identified theme(s) without using qualitative descriptors (e.g., 'good', 'poor'). 
        3. Do not output any explanatory comments such as: 'Here are the main themes...'. 
        4. provide a theme if and only if it is mentioned by at least '5%' of the responses. 
        5. Try to be concise and avoid redundancy. """
    chunk_indices = [0]
    responses_string = ""
    index_response = 0
    free_space = 20
    number_themes =   10
    len_theme = 20
    len_output =  number_themes*len_theme + free_space
    len_instruction = len(tokenizer.encode(system_prompt))
    cost = 0
    for index, response in enumerate(responses):
        responses_string =  responses_string + str((index_response+1)) + "." + response + "\n" 
        len_responses = len(tokenizer.encode(responses_string))
        if (len_responses + len_instruction + len_output) >= context_length:
            chunk_indices.append(index)
            index_response = 0
            responses_string = str((index_response+1)) + "." + response + "\n"
            cost+= (len_responses + len_instruction)*per_token_input_cost/1000 + len_output*per_token_output_cost/1000       
    if chunk_indices[-1] != len(responses):
        chunk_indices.append(len(responses))
        cost+= (len_responses + len_instruction)*per_token_input_cost/1000 + len_output*per_token_output_cost/1000
    return (chunk_indices, cost)

def calculate_chunks_response_mapping(responses, themes, system_prompt, model):

    context_length = model.context_length
    per_token_input_cost = model.input_cost
    per_token_output_cost = model.output_cost
    if (system_prompt == ""):
        system_prompt = """
        You have identified the main themes in the open-ended provided responses that explain the respondents'
        multiple choice selection. 
        Please assign each response to one or more themes from the list of themes provided. 
        Instructions:
        1. Do not output any explanatory comments such as: "Here are the themes assigned...".
        2. Do not add any additional themes to the list provided and just pick one or more from the list.
        3. Assign a row to each response with the theme(s) mapped to it according to the output example.
        4. If a response does not fit any of the themes provided, leave the cell blank.
        5. Output: Provide the result in the specified json schema: 
        {"mapped_responses": [
        {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id},
        {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id},
        ...
        {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id}
        ]}"""
    themes_string = ",".join([str(id) + '. ' + str(theme) + '\n' for id, theme in enumerate(themes)])
    len_mapped_responses_format = len(tokenizer.encode("""{"theme(s)":[], "response_id": id}""")) + len(tokenizer.encode("theme_id,")) * len(themes)
    len_prefix_response = len(tokenizer.encode("""{"mapped_response":[]}"""))
    chunk_indices = [0]
    responses_string = ""
    index_response = 0
    cost = 0
    len_instruction = len(tokenizer.encode(system_prompt)) + len(tokenizer.encode("responses: , themes:")) + len(tokenizer.encode(themes_string))
    for index, response in enumerate(responses):
        responses_string =  responses_string + str((index_response+1)) + "." + response + "\n" 
        len_output = len_mapped_responses_format*(index_response+1) + len_prefix_response
        len_responses = len(tokenizer.encode(responses_string))
        if (len_responses + len_instruction + len_output) >= context_length:
            chunk_indices.append(index)
            index_response = 0
            responses_string = str((index_response+1)) + "." + response + "\n"  
            cost+= (len_responses + len_instruction)*per_token_input_cost/1000 + len_output*per_token_output_cost/1000     
    if chunk_indices[-1] != len(responses):
        chunk_indices.append(len(responses))
        cost+= (len_responses + len_instruction)*per_token_input_cost/1000 + len_output*per_token_output_cost/1000
    return (chunk_indices, cost)



import pandas as pd
import numpy as np
import json
import subprocess
from LLM_toolkit import create_llm
import argparse
import os
import ast
import yaml
import re

# Ideally this should be a function that gets run (or is called directly) after the high level pipeline function

# By question, the input is question text, extracted insights, and representative responses (embeddings?)

# Save the output to a J-SON file that can be incorporated into an R markdown template which is called at the end
# Can discuss this bit, but this would allow us to use the package Harry shared 


def load_summary_info(input_path,s_q_map_path):
    """
    Load in the summary data, return all the information in dictionaries with question vairbales as the keys.

    Parameters:
    - input_path (str, optional): Path to the output file in the high level pipeline.

    Returns:
    - thematically_coded_dict: A dictionary (keys are the quesitons) of themetically coded response data, values are the dataframe outputs from the high level pipeline.
    - insight_dict: A dictionary (keys are the quesitons) of extracted insights, values are the dataframe outputs from the high level pipeline.
    """
    # get the thematically coded responses (one sheet per dict)
    thematically_coded_dict = pd.read_excel(input_path, sheet_name=None)

    # get the cluster summaries file_name
    file_name = input_path.replace(".xlsx", "_cluster_summaries.xlsx")
    # read in all the sheets to a dictionary (one sheet per dict)
    insight_dict = pd.read_excel(file_name, sheet_name=None)

    # read in all the questions, convert to a dict
    question_text = pd.read_excel(s_q_map_path)
    question_dict = question_text.set_index('question_column_names')['machine_readable_question_text'].to_dict()

    return {"thematically_coded_dict":thematically_coded_dict, 
            "insight_dict":insight_dict, 
            "question_dict":question_dict}

def save_written_summaries(written_summary_dict,
                           input_path):
    """
    Saves the generated summary information to JSON.

    Parameters:
    - written_summary_dict (dict): The LLM generated summary dict.
    - input_path (str): Path to the folder where the output should be stored.

    Returns:
    - None: Save the summary information in json format.
    """

    file_name = input_path.replace(".xlsx", "_llm_summaries.json")

    with open(file_name, 'w') as f:
        json.dump(written_summary_dict, f)


def create_written_summaries(model,
                             insight,
                             thematically_coded_dict,
                             question_text,
                             col):
    """
    Orchestrates the process of analyzing high-level data using configurations specified in an INI file.

    Parameters:
    - model: An LLM object that can be used to generate summaries.
    - insight (str): The insight to be summarised.
    - thematically_coded_dict (df): A dataframe of the responses for the question this insight is derived from.
    - question_text (str): The question text of the insight.
    - col (str, optional): The column name the insight is coming from (should have the MCQ value if applicable)

    Returns:
    - Dictionary: A dictionary containing all the information need to produce a summary of the provided template.
    """

    mcq_response = False # keep this false for now as we figure out the best way to incorporate multiple choice selections

    # filter to rows to responses containg the theme
    try:
        insight_df =  thematically_coded_dict[thematically_coded_dict["extracted_insights_aggregate"].str.contains(insight)]
    except TypeError:
        print(f"Issue with Insight: {insight}; returning an empty string")
        return ""
    
    # If the insight is from an mcq it is in a different column...
    if len(insight_df)==0:
        insight_df =  thematically_coded_dict[thematically_coded_dict["extracted_insights_by_mcq"].str.contains(insight)]
        mcq_response = True
        print("using the MCQ column")

    num_supporters = len(insight_df)
    representative_responses = insight_df.sample((5 if len(insight_df)>5 else len(insight_df)))["response"].tolist()

    
    join_character = "\n"
    user_prompt = f"""
    The insight "{insight}" was extracted through a thematic analysis of the question: "{question_text}".
    In a short pargraph, your task is to summarise the reasoning behind this extracted insight drawing on the context of the representative responses provided below. 
    {f'For additional context, keep in mind that this insight is from the subset of respondents which selected "{col}" on a related multiple choice question. ' if mcq_response else ""}
    1. Respond as succinctly as possible, while providing enough information to give a policy analyst a good understanding of the extracted insight.
    2. Do not directly quote any of the representative responses or make up information that does not follow logically from the representative responses. 
    3. Do not include any explanatory statements such as "The summary of this insight is" or "Given the context of the representative responses". 
    4. Provide only the short summary paragraph as free text, without additional explanatory information after the summary is returned. 
    5. Ensure the summary sticks to only explaining the reasoning behind the extracted insight "{insight}" without straying into other topic areas.
    6. If you are unable to generate a summary due to a lack of context, please respond with "Not enough information to summarise this insight". 
    Representative responses which mentioned this extracted insight:
    {join_character.join(["R"+str(a+1)+": "+b for a,b in enumerate(representative_responses)])}
    """

    # Generate summary for the current cluster using the language model
    llm_summary = model.run(user_prompt)

    return {"short_summary":llm_summary,
            "num_supporters":num_supporters,
            "representative_responses":representative_responses}


def create_final_templates(input_path,consultation_name,rscript_location):
    """
    Executes R code to generate all of the summary output files.

    Parameters:
    - input_path (str, optional): Path to the folder where the output should be stored.

    Returns:
    - None: Executes R code to generate all of the summary output files.
    """
    json_file = input_path.replace(".xlsx", "_llm_summaries.json")
    output_directory = os.path.dirname(input_path)

    subprocess.call([rscript_location,'src/create_prompt_templates.R', json_file, output_directory,consultation_name])




def generate_written_summaries(config_path):
    """
    Orchestrates the process of generating written summaries, after the high level pipeline is run.

    Parameters:
    - directory_path (str, optional): Path to the directory where the pipeline output is saved.
    - model (str, optional): An LLM-factory object that can be used to generate summaries.

    Returns:
    - None: Executes the full pipeline from reading in data, extracting insights, and saving results.
    """
    # First load in and organize the data that we will need
    
    with open(config_path, 'r') as file:
            config = yaml.safe_load(file)
    
    consultation_name = config['GENERAL']['consultation_name']
    output_dtm_str = config['LOW_LEVEL']['FILES']['output_datetime']
    
    input_path = os.path.join('data', consultation_name, 'output', output_dtm_str, 'high_level_analysis_output.xlsx')
    s_q_map_path = os.path.join('data', consultation_name, 'input', config['GENERAL']['sheet_question_mapping_file'])
    rscript_location = config['R']['rscript_location']
    model = create_llm(**dict(config['HIGH_LEVEL']['MODEL']))

    
    
    print('Loading summary info...')
    summary_info = load_summary_info(input_path,
                                     s_q_map_path)
    

    # Use a LLM to generate the written summaries
    print("Creating written summaries...")
    written_summary_dict = {}
    for question in summary_info["insight_dict"]: # iterate over questions
        print(f"For question: {question} which has {len( [a for a in list(summary_info['insight_dict'][question].columns) if 'responses' not in a])} question columns...")
        question_mcq_summary_dict = {}
        for col in [a for a in list(summary_info["insight_dict"][question].columns) if "responses" not in a]: # iterate over insight columns (aggregate + any others)
            question_summary_dict = {}
            for insight in summary_info["insight_dict"][question][col].tolist():
                temp_summary_dict = create_written_summaries(model,
                                                             insight,
                                                             summary_info["thematically_coded_dict"][question],
                                                             summary_info["question_dict"][question],
                                                             col)
                if temp_summary_dict=="":
                    continue
                # add in the top n responses
                top_n_responses = summary_info["insight_dict"][question].loc[summary_info["insight_dict"][question][col] == insight, col+"_repr_responses"].iloc[0]
                # Having trouble with this formatting, converting a list
                top_n_responses = ast.literal_eval(top_n_responses)
                temp_summary_dict["top_n_responses"] = [a['response'] for a in top_n_responses]
                temp_summary_dict["top_n_low_level_themes"] = [a['low_level_themes_used'] for a in top_n_responses]
                question_summary_dict[insight] = temp_summary_dict
            
            question_mcq_summary_dict[col]=question_summary_dict
        written_summary_dict[question]={"question_text":summary_info["question_dict"][question],
                                        "summary_dict":question_mcq_summary_dict,
                                        "total_responses":len(summary_info["thematically_coded_dict"][question])}


    # Save the output to a J-SON file that can be incorporated into an R markdown template
    print("Saving written summaries...")
    save_written_summaries(written_summary_dict,
                           input_path)

    # Create final templates (call the R code)
    print("Creating creating the final written output documents...")
    create_final_templates(input_path,consultation_name,rscript_location)
    print("Output file created sucessfully.")


import os
import pandas as pd
import re
def read_input_data(consultation_name, model_name, demo, mode = 'theme_extraction'):
    # output and input paths
    input_path = os.path.join('data', consultation_name, 'input', 'input_file.xlsx')
    s_q_map_path = os.path.join('data', consultation_name, 'input', 'sheet_question_mapping.xlsx')
    xls = pd.ExcelFile(input_path)
    sheet_names = xls.sheet_names
    dataframes = {f'{name}': pd.read_excel(xls, sheet_name=name) for name in sheet_names}
    
    dataframes['df_q1'].loc[dataframes['df_q1']['mcq_response'] == 'Strongly agree', 'mcq_response'] = 'Agree'
    dataframes['df_q1'].loc[dataframes['df_q1']['mcq_response'] == 'Strongly disagree', 'mcq_response'] = 'Disagree'
       
    # Check if all dataframes first column is id
    for key, df in dataframes.items():
        if df.columns[0] != 'id':
            raise ValueError(f'First column of dataframe {key} is not "id"')
    
    # Rename the second column in each dataframe to 'response', and the third column (if it exists) to 'mcq_response'
    for key, df in dataframes.items():
        # Check n of unique values in the mcq_response column
        if 'mcq_response' in df.columns:
            if len(df['mcq_response'].unique()) > 10:
                raise ValueError(f'Number of unique values in the mcq_response column of df {key} exceeds 10. Please check the data to make sure this in fact is the appopriate column.')

    # If demo is True, only extract low-level insights from the first 10 rows of each dataframe
    if demo['boolean']==True:
        for key, df in dataframes.items():
            df = df.sample(demo['n_examples'])
            df.sort_values(by='id', inplace=True)
            df.reset_index(drop=True, inplace=True)
            dataframes[key] = df
    
    # Read in the sheet-question mapping (to be included in the prompt)
    s_q_map = pd.read_excel(s_q_map_path)
    s_q_map_dict = {key: value for key, value in s_q_map.iloc[:, :2].values}

    # Create the output directory and path
    themes = {}
    if mode != 'theme_extraction':
        try:
            themes_input_path = os.path.join('data', f'{consultation_name}/output/insight_{model_name}.xlsx')
            xls = pd.ExcelFile(themes_input_path)
            sheet_names = xls.sheet_names
            themes = {f'{name}': pd.read_excel(xls, sheet_name=name) for name in sheet_names}
        except:
            print('No themes found for this model') 
    return dataframes, themes, s_q_map_dict                                               

def clean_gpt_3_output():
     ## make the format of all equal? like strip for example?
    model = 'gpt-35-turbo-16k' ## clean the insight file 
    consultation_names = ['turing_consultation', 'synthetic_consultation']
    for consultation_name in consultation_names:
            output_dir = os.path.join('data', 'Brute_force')
            insight_path = os.path.join(output_dir, f'2024-09-08_{consultation_name}/insight_gpt-35-turbo-16k.xlsx')
            insight_xlsx = pd.ExcelFile(insight_path)
            responses_writer = pd.ExcelWriter(insight_path.replace('insight','clean_insight'))
            sheet_names = insight_xlsx.sheet_names
            insight_dataframes = {f'{name}': pd.read_excel(insight_xlsx, sheet_name=name) for name in sheet_names}
            for df_name in insight_dataframes.keys():
                insight_df = insight_dataframes[df_name]
                for index, row in insight_df.iterrows():
                    themes_list = re.findall(r"'([^']*)'", row['themes1'])
                    for theme in themes_list:
                        insight_df.loc[index, f'themes{themes_list.index(theme)+1}'] = theme
                insight_df.to_excel(responses_writer, sheet_name=df_name, index=False)
            responses_writer.close()
    return None
def clean_output():
    # output and input paths
    consultation_names = ['turing_consultation', 'synthetic_consultation']
    models = ['gpt-4','gpt-4o','gpt-4o-mini', 'gpt-35-turbo-16k', 'llama-3-8b-instruct','llama-3-70b-instruct', 'llama-2-70b-chat']
    llamas = ['llama-2-70b-chat']
    for model in llamas:
        for consultation_name in consultation_names:
            output_dir = os.path.join('data', 'Brute_force')
            insight_path = os.path.join(output_dir, f'2024-09-08_{consultation_name}/insight_{model}.xlsx')
            response_path = os.path.join(output_dir, consultation_name, f'response_result_{model}.xlsx')
            responses_writer = pd.ExcelWriter(response_path.replace('response_result','clean_response'))
            insight_xlsx = pd.ExcelFile(insight_path)
            response_xlsx = pd.ExcelFile(response_path)
            sheet_names = response_xlsx.sheet_names
            insight_dataframes = {f'{name}': pd.read_excel(insight_xlsx, sheet_name=name) for name in sheet_names}
            response_dataframes = {f'{name}': pd.read_excel(response_xlsx, sheet_name=name) for name in sheet_names}
            print(consultation_name, model)
            for df_name in insight_dataframes.keys():
                insight_df = insight_dataframes[df_name]
                response_df = response_dataframes[df_name]
                choices = insight_df['choice']
                theme_columns = [column for column in insight_df.columns if 'theme' in column]
                for choice in choices:
                    insight_dict = insight_df.loc[insight_df['choice']==choice ,theme_columns].to_dict(orient='records')[0]
                    insight_dict = {key.replace('themes', 'theme_'):value.replace('.','') for key, value in insight_dict.items() if not pd.isnull(value)}
                    a = response_df.loc[response_df['mcq_response']==choice, ['theme(s)','id']]
                    for index, row in a.iterrows():
                        if not pd.isnull(row['theme(s)']):
                            for key, value in insight_dict.items():
                                value = value.strip().replace("'",'')
                                row['theme(s)'] = row['theme(s)'].replace(value, f'{key}')
                            if 'theme_' not in row['theme(s)']:
                                row['theme(s)'] = re.sub(r'\b\d+\b', lambda x: f'theme_{x.group()}', row['theme(s)'])
                            row['theme(s)'] = row['theme(s)'].replace('\'','')
                            row['theme(s)'] = row['theme(s)'].replace('"','')
                            if '[' not in row['theme(s)']:
                                row['theme(s)'] = '[' + row['theme(s)'] + ']'
                            response_df.loc[response_df['id']==row['id'], 'theme(s)'] = str(row['theme(s)'])
                response_df.to_excel(responses_writer, sheet_name=df_name, index=False)
            responses_writer.close()
if __name__ == "__main__":
    clean_output()            


import pandas as pd
from typing import Dict, Tuple
from sklearn.cluster import KMeans
import json
from litellm import embedding
from output_summaries import generate_written_summaries
from datetime import datetime
from copy import deepcopy
from LLM_toolkit import create_llm
from pydantic import ValidationError
from typing import Dict
from pydantic import BaseModel, validator
from typing import List
from transformers import GPT2Tokenizer
from cost import calculate_chunks_response_mapping
import os
import re
import tiktoken
tokenizer = tiktoken.encoding_for_model('gpt-3.5-turbo')
from data_cleaner import read_input_data

class LowLevelTheme(BaseModel):
    theme: str

class LowLevelResponseAnalysis(BaseModel):
    themes: List[LowLevelTheme]

def map_response_chunks(model, prompt):
    mapping_df = pd.DataFrame(columns= ['response','theme(s)'])
    if ('gpt-4' in model.model):
        llm_response = model.run(user_prompt = prompt , response_format={"type": "json_object"})
        llm_response = json.loads(llm_response)
        for mapped_response in llm_response['mapped_responses']:
            mapping_df = pd.concat([mapping_df, pd.DataFrame([mapped_response])], axis=0)
        return mapping_df
    else:
        llm_response = model.run(user_prompt = prompt)
        llm_response = llm_response.split("mapped_responses")[1]
        mapped_responses = re.split(r'{*},', llm_response)
        for i, mapped_response in enumerate(mapped_responses):
            if "theme(s)" in mapped_response and "response_id" in mapped_response:
                themes = mapped_response.split('theme(s)')[1].split("[")[1].split("]")[0]
                response_id = int(re.findall(r'\d+', mapped_response.split('"response_id"')[1])[0])
                mapping_df = pd.concat([mapping_df, pd.DataFrame([{'response_id': response_id, 'theme(s)': themes}])], axis=0)
        return mapping_df
    
def map_responses_for_question(df: pd.DataFrame, themes, question_string, model) -> pd.DataFrame:
    prompt_instruction = """
    You have identified the main themes in the open-ended provided responses that explain the respondents'
    multiple choice selection. 
    Please assign each response to one or more themes from the list of themes provided. 
    Instructions:
    1. Do not output any explanatory comments such as: "Here are the themes assigned...".
    2. Do not add any additional themes to the list provided and just pick one or more from the list.
    3. Assign a row to each response with the theme(s) mapped to it according to the output example.
    4. If a response does not fit any of the themes provided, leave the cell blank.
    5. Output: Provide the result in the specified json schema: 
    {"mapped_responses": [
    {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id},
    {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id},
    ...
    {"theme(s)": [theme_id, theme_id,...,theme_id], "response_id": id}
    ]}"""
    choices = df['mcq_response'].unique()
    df['response_id'] = df.index
    df['theme(s)'] = None
    theme_columns = [theme_column for theme_column in themes.columns if 'theme' in theme_column]
    for choice in choices:
        responses = df.loc[df['mcq_response'] == choice, 'response']
        ids = df.index[df['mcq_response'] == choice].tolist()
        prompt_question = f"""Please analyze the following survey responses and map responses to the themes:
        Question: {question_string}
        {"Multiple Choice Selection: " + choice}"""
        choice_themes = themes.loc[themes['choice'] == choice, [theme_column for theme_column in theme_columns]].values.tolist()[0]
        themes_string = ""
        for i,theme in enumerate(choice_themes):
            if theme != None:
                themes_string =  themes_string + f'theme_{i+1}: {theme}\n'
        chunk_indices, cost = calculate_chunks_response_mapping(responses, choice_themes, prompt_question + prompt_instruction, model)
        num_chunks = len(chunk_indices)-1
        print(f"mapping responses for : {choice} with number of chunks: {num_chunks} and number of responses: {len(responses)}")
        for index in range(len(chunk_indices)-1): 
            response_chunk = responses[chunk_indices[index]:chunk_indices[index+1]] 
            id_chunks = ids[chunk_indices[index]:chunk_indices[index+1]]
            response_string = ""
            for i,response in enumerate(response_chunk):
                response_string =  response_string + f'response_id: {id_chunks[i]}, response_text: {response}\n'
            prompt_input = f"""Open-ended Responses: {response_string} 
            themes : {themes_string} """
            prompt = prompt_question + prompt_input + prompt_instruction
            print(f"analyzing chunk {index} with len {len(tokenizer.encode(prompt))} and number of responses: {len(response_chunk)}")
            llm_response = map_response_chunks(model, prompt)
            print('llm responded')
            df = df.merge(llm_response[['response_id','theme(s)']], on='response_id', how='left')
            df['theme(s)'] = df['theme(s)_x'].combine_first(df['theme(s)_y'])
            df.drop(columns=['theme(s)_x', 'theme(s)_y'], inplace=True)
    df.drop(columns=['response_id'], inplace=True)
    return df, cost

def response_Brute_force_pipeline(demo, consultation_name, model_name): 
    
    ### To separate context length later

    consultation_data, themes, s_q_map_dict = read_input_data(consultation_name, model_name, demo, 'response_mapping')
    
    output_path = os.path.join('data', f'{consultation_name}/output/response_{model_name}.xlsx')
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    responses_writer = pd.ExcelWriter(output_path) 
    
    ## This should be somewhere else for the final version
    model = create_llm(model_name)
    cost_df = 0
    for consultation_key, responses in consultation_data.items():
        current_question = s_q_map_dict[consultation_key]
        print(f"Analyzing {consultation_key}")
        mapped_responses,cost = map_responses_for_question(responses, themes[consultation_key], current_question, model)
        mapped_responses.to_excel(responses_writer, sheet_name= consultation_key, index=False)
        cost_df += cost
    print(f"Total cost of the analysis {cost_df}")
    responses_writer.close()




import os
from typing import List, Dict, Any
import requests
import time
from litellm import completion, InternalServerError
from dotenv import load_dotenv
import yaml  # Add this import
import json

class LLM_Base:
    """
    Abstract base class for different LLM types.
    """
    def __init__(self, model: str, api_key: str):
        self.model = model
        self.api_key = api_key

    def run(self, user_prompt:str, system_prompt:str="You are a helpful assistant.", **kwargs: Any) -> str:
        messages = [
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt}
                    ]
        max_retries = 12
        for attempt in range(max_retries):
            try:
                if 'gpt' in self.model:
                    return self._completion(messages, **kwargs)
                else:
                    llm_inputs = {
                    "messages": [
                        {"role": "system", "content": system_prompt},
                        {"role": "user", "content": user_prompt}
                    ],
                    **kwargs
                    }
                    self.headers = {
                    'Authorization': f'Bearer {self.api_key}',
                    'Content-Type': 'application/json'
                    }
                    response = requests.post(self.endpoint, headers=self.headers, data=json.dumps(llm_inputs))
                    return response.json()['choices'][0]['message']['content']
            except (requests.ConnectionError, KeyError, AttributeError, InternalServerError) as e:
                if attempt < max_retries - 1:
                    print(f"{e}, retrying...")
                    time.sleep(5)  # Wait for 5 seconds before retrying
                    continue
                else:
                    raise  # Re-raise the last exception if all retries fail

    def _completion(self, messages: List[Dict[str, str]], **kwargs: Any) -> str:
        raise NotImplementedError("Subclass must implement this method")


class LITELLM(LLM_Base):
    def __init__(self,model_config):
        super().__init__(model_config['model'], model_config['api_key'])
        self.api_type = "litellm"
        self.context_length = model_config['context_length']

    def _completion(self, messages: List[Dict[str, str]], **kwargs: Any) -> str:
        chat_completion = completion(
            model=self.model,
            api_key=self.api_key,
            messages=messages,
            **kwargs
        )
        return chat_completion['choices'][0]['message']['content']


class LITELLM_AZURE(LLM_Base):
    def __init__(self, model_config):
        super().__init__(model_config['model'], model_config['api_key'])
        self.api_version = model_config['api_version']
        self.endpoint = model_config['endpoint']
        self.api_type = "azure"
        self.context_length = model_config['context_length']
        self.input_cost = model_config['input_cost']
        self.output_cost = model_config ['output_cost']

    def _completion(self, messages: List[Dict[str, str]], **kwargs: Any) -> str:
        chat_completion = completion(
            model=f"azure/{self.model}",
            api_base=self.endpoint,
            api_version=self.api_version,
            api_key=self.api_key,
            messages=messages,
            **kwargs
        )
        return chat_completion['choices'][0]['message']['content']


def get_model_config(model_name: str) -> Dict[str, str]:
    model_configs_path = os.path.join('src', 'config', 'model_configs.yaml')
    with open(model_configs_path, 'r') as file:
        configs = yaml.safe_load(file)
    
    if model_name not in configs['LLMs']:
        raise ValueError(f"Unknown model: {model_name}")
    
    return configs['LLMs'][model_name]

def create_llm(model_name: str) -> LLM_Base:
    model_config = get_model_config(model_name)
    if model_config['api_type'] == 'litellm':
        return LITELLM(model_config)
    elif model_config['api_type'] == 'litellm_azure':
        return LITELLM_AZURE(model_config)
    else:
        raise ValueError(f"Unknown model type for {model_name}")


from setuptools import setup, find_packages
setup(
    name="MIMDE",
    version="1.0.4",
    packages=find_packages(),
    install_requires=[],  # List your dependencies
    description="Python package for MIMDE",
    url="https://github.com/ai-for-public-services/MIMDE",
    authors="Saba Esnaashari, John Francis, and Anton Appolonov",
    author_email="sesnaashari@turing.ac.uk",
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',        # Minimum Python version required
    # Add additional metadata for publishing
    keywords="MIMDE, package, python",
    project_urls={
        "Bug Tracker": "https://github.com/ai-for-public-services/MIMDE/issues",
        "Documentation": "https://github.com/ai-for-public-services/MIMDE/wiki",
        "Source Code": "https://github.com/ai-for-public-services/MIMDE",
    },
)


import pandas as pd
from typing import Dict, Tuple
from sklearn.cluster import KMeans
import json
from litellm import embedding
from output_summaries import generate_written_summaries
from datetime import datetime
from copy import deepcopy
from .LLM_toolkit import create_llm
from pydantic import ValidationError
from typing import Dict
from pydantic import BaseModel, validator
from typing import List
from transformers import GPT2Tokenizer
import os
import tiktoken
from pydantic_classes import LowLevelResponseAnalysis, RoughHighLevelAnalysis, HighLevelResponseAnalysis
tokenizer = tiktoken.encoding_for_model('gpt-3.5-turbo')
#tokenizer = T5Tokenizer.from_pretrained("t5-small")
import re
from .data_cleaner import read_input_data
from cost import calculate_chunks_theme_extraction
class LowLevelTheme(BaseModel):
    theme: str

class LowLevelResponseAnalysis(BaseModel):
    themes: List[LowLevelTheme]


def json_format(llm_response, PydanticClass):
    """
    Attempts to generate a valid final themes JSON response using the specified model.
    
    Notee: This function is called by 'identify_themes'.
    """
    # Attempt to generate a valid final themes JSON
    for attempt in range(5):
        try:
            llm_response = json.loads(llm_response)
            error = None
            return llm_response['themes']
        except (json.decoder.JSONDecodeError, ValueError, ValidationError) as e:
            if attempt < 4:
                print(f"Attempt {attempt+1} to generate a valid final themes JSON failed due to a {e}, retrying...")
                continue
            else:
                if PydanticClass==LowLevelResponseAnalysis:
                    print(f"Attempt {attempt+1} to generate a valid final themes JSON failed due to a {e}, retrying...")
                    print("Using default error schema for LowLevelResponseAnalysis...")
                    # Default error schema for LowLevelResponseAnalysis
                    default_error_schema = {
                        "themes": [{"theme": "Error: Unable to generate valid themes"}],
                        "mcq_contradiction": False,
                        "outlier": {
                            "is_outlier": False,
                            "outlier_category": {
                                "irrelevant": False,
                                "incoherent": False,
                                "extreme": False,
                                "other": False
                            },
                            "outlier_reason": None,
                            "outlier_score": 0.0
                        }
                    }
                    llm_response = PydanticClass(**default_error_schema)
                    return llm_response
                else:
                    print(f"Attempt {attempt+1} to generate a valid final themes JSON failed due to a {e}, please check the prompt and try again.")
                    raise
def ask_llm(model, user_prompt):
    # Attempt to generate a valid final themes JSON  - work on this a little more later
    if ('gpt-4' in model.model):
       llm_response = model.run(user_prompt = user_prompt, response_format={"type": "json_object"}) 
       return json_format(llm_response, LowLevelResponseAnalysis)
    else:
        for attempt in range(5):
            llm_response = model.run(user_prompt = user_prompt)
            try:
                themes_string = llm_response.split("[")[1].split("]")[0]
                themes_list = re.split(r'\s*,\s*', themes_string)
                return themes_list
            except:
                print(f"unable to extract themes in attempt {attempt+1}\nllm response: {llm_response}")
                continue
        return []

def identify_themes(df: pd.DataFrame, model, question_string:str, prompt_instruction:str = "") -> Tuple[pd.DataFrame, pd.DataFrame]:

    themes_df = pd.DataFrame()
    choices = df['mcq_response'].unique()
    total_cost = 0
    for choice in choices:
        responses = df.loc[df['mcq_response'] == choice, 'response']
        prompt_input = f"""Please analyze the following survey responses:
            Question: {question_string}
            "Multiple Choice Selection: " {choice}
            Open-ended Responses: """
        chunk_indices, cost = calculate_chunks_theme_extraction(responses, prompt_input+prompt_instruction, model)
        total_cost += cost
        num_chunks = len(chunk_indices)-1
        print(f"Extracting themes for: {choice} with number of chunks: {num_chunks} and number of responses: {len(responses)}")
        llm_response_list = []
        for index in range(len(chunk_indices)-1): 
            response_chunk = responses[chunk_indices[index]:chunk_indices[index+1]] 
            responses_string = ""
            for i, response in enumerate(response_chunk):
                responses_string =  responses_string + str((i+1)) + "." + response + "\n" 
            prompt = prompt_input + responses_string + prompt_instruction
            print(f"analyzying chunk {index} with len {len(tokenizer.encode(prompt))} and number of responses: {len(response_chunk)}")
            llm_response = ask_llm(model, prompt)
            llm_response_list = llm_response_list + llm_response
        if num_chunks > 1:
            prompt =  f""" please analyze the following survey themes found and extract the main themes from the list provided.
            Question: {question_string}
            "Multiple Choice Selection: " {choice}
            Themes: {llm_response_list}""" + prompt_instruction.split("4.")[0] + "Output: Provide the theme(s)in the specified list {{'themes':[ str, str, ..., str], }}"
            print(f"analyzing all chunks with len {len(tokenizer.encode(prompt))}") 
            llm_response = ask_llm(model, prompt)
        themes = {}        
        for i, theme in enumerate(llm_response):
            themes[f'themes{i+1}'] = theme
        themes['number_of_responses'] = len(responses)
        themes['choice'] = choice
        themes_dict = pd.DataFrame([themes])
        themes_df = pd.concat([themes_df, themes_dict], axis=0)
        print(f"Analyzing responses for: {choice}") 
    return themes_df, total_cost

def themes_Brute_force_pipeline(demo, consultation_name, model_name, prompt_instruction:str = ""): 
    prompt_instruction = """Instructions: 
        1. Identify the main theme(s) in the open-ended provided responses that explain the respondents' multiple choice selection. 
        2. Provide detailed identified theme(s) without using qualitative descriptors (e.g., 'good', 'poor'). 
        3. Do not output any explanatory comments such as: 'Here are the main themes...'. 
        4. provide a theme if and only if it is mentioned by at least '5%' of the responses. 
        5. Try to be concise and avoid redundancy. """
    list_output_format =  """ Output: Provide the theme(s)in the specified list {{'themes':[ str, str, ..., str], }}"""
    json_output_format = """Output: Provide the theme(s)in the specified json schema: {{ "themes": [str, strs, ..., str]}}"""
    if 'gpt-4' in model_name:
        prompt_instruction = prompt_instruction + json_output_format
    elif 'llama' in model_name:
        prompt_instruction = prompt_instruction.split("5.")[0] + json_output_format
    else:
        prompt_instruction = prompt_instruction + list_output_format

    consultation_data, themes, s_q_map_dict = read_input_data(consultation_name, model_name, demo, 'theme_extraction')
    output_path = os.path.join('data', f'{consultation_name}/output/insight_{model_name}.xlsx')
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    themes_writer = pd.ExcelWriter(output_path)
    model = create_llm(model_name)

    # Iterate over the dataframes and extract insights for each question
    dataframes_annotated = {}
    total_cost = 0
    for key, df in consultation_data.items():
        current_question = s_q_map_dict[key]
        print(f"Analyzing {key}")
        themes,cost = identify_themes(df, model, question_string=current_question, prompt_instruction=prompt_instruction)
        total_cost += cost
        dataframes_annotated[key] = themes
        themes.to_excel(themes_writer, sheet_name=key, index=False)
    print(f"Total cost for theme extraction: {total_cost}")
    themes_writer.close()



import json
from typing import List, Optional, Dict
from pydantic import BaseModel, validator


#------------------------ Low Level Analysis ------------------------#
class LowLevelTheme(BaseModel):
    theme: str

class OutlierCategory(BaseModel):
    irrelevant: bool
    incoherent: bool
    extreme: bool
    other: bool

class Outlier(BaseModel):
    is_outlier: bool
    outlier_category: OutlierCategory
    outlier_reason: Optional[str] = None
    outlier_score: Optional[float] = 0.0
    # Set outlier_score to 0.0 if it is None or not a number
    @validator('outlier_score', pre=True)
    def set_outlier_score(cls, v):
        if v is None or not isinstance(v, (int, float)):
            return 0.0
        return float(v)

class LowLevelResponseAnalysis(BaseModel):
    themes: List[LowLevelTheme]
    mcq_contradiction: bool = False
    outlier: Outlier
#-------------------------------------------------------------------#



#--------------------- Rough High Level Analysis -------------------#
class RoughHighLevelTheme(BaseModel):
    theme: str

class RoughHighLevelAnalysis(BaseModel):
    themes: List[RoughHighLevelTheme]
#-------------------------------------------------------------------#



#----------------------- High Level Analysis -----------------------#

class HighLevelTheme(BaseModel):
    cluster_n: int
    theme: str
    # Set validator to ensure that cluster_n is an integer and transform a str to an int
    @validator('cluster_n')
    def cluster_n_must_be_int(cls, v):
        if not isinstance(v, int):
            return int(v)
        return v

class HighLevelResponseAnalysis(BaseModel):
    themes: List[HighLevelTheme]
